@@ -156,6 +156,7 @@ approval_free (Approval *approval)
     /* we should have replied to the method call by now */
     g_assert (approval->context == NULL);
 
+    g_free (approval->client_bus_name);
     g_slice_free (Approval, approval);
 }
 
@@ -243,6 +244,13 @@ struct _McdDispatchOperationPrivate
      * A reference is held for each pending approver. */
     gsize ado_pending;
 
+    /* The number of plugins whose decision we're waiting for,
+     * regarding whether a handler is in fact suitable. */
+    gsize handler_suitable_pending;
+
+    /* If non-NULL, a plugin has decided the selected handler is unsuitable. */
+    GError *handler_unsuitable;
+
     /* If TRUE, we're dispatching a channel request and it was cancelled */
     gboolean cancelled;
 
@@ -250,9 +258,9 @@ struct _McdDispatchOperationPrivate
      * after observers */
     gboolean observe_only;
 
-    /* If TRUE, we're in the middle of calling HandleChannels. This is a
-     * client lock. */
-    gboolean calling_handle_channels;
+    /* If non-NULL, we're in the middle of asking plugins whether we may call
+     * HandleChannels, or doing so. This is a client lock. */
+    McdClientProxy *trying_handler;
 
     /* If TRUE, we've tried all the BypassApproval handlers, which happens
      * before we run approvers. */
@@ -457,9 +465,9 @@ _mcd_dispatch_operation_check_client_locks (McdDispatchOperation *self)
 
     /* if we've called one Handler, we may not continue until it responds
      * with an error */
-    if (self->priv->calling_handle_channels)
+    if (self->priv->trying_handler != NULL)
     {
-        DEBUG ("waiting for HandleChannels to return");
+        DEBUG ("waiting for handler_is_suitable or HandleChannels to return");
         return;
     }
 
@@ -494,6 +502,10 @@ _mcd_dispatch_operation_check_client_locks (McdDispatchOperation *self)
                 continue;
 
             DEBUG ("Internal handler for request channel #%u", i);
+            _mcd_handler_map_set_channel_handled_internally (
+                self->priv->handler_map,
+                mcd_channel_get_tp_channel (channel),
+                _mcd_dispatch_operation_get_account_path (self));
             _mcd_request_handle_internally (request, channel, TRUE);
         }
 
@@ -503,6 +515,18 @@ _mcd_dispatch_operation_check_client_locks (McdDispatchOperation *self)
         return;
     }
 
+    /* If there are no potential handlers, the story ends here: we don't
+     * want to run approvers in this case */
+    if (self->priv->possible_handlers == NULL)
+    {
+        GError incapable = { TP_ERRORS, TP_ERROR_NOT_CAPABLE,
+            "No possible handlers, giving up" };
+
+        DEBUG ("%s", incapable.message);
+        _mcd_dispatch_operation_close_as_undispatchable (self, &incapable);
+        return;
+    }
+
     approval = g_queue_peek_head (self->priv->approvals);
 
     /* if we've been claimed, respond, then do not call HandleChannels */
@@ -543,6 +567,7 @@ _mcd_dispatch_operation_check_client_locks (McdDispatchOperation *self)
             g_source_remove (approver_event_id);
         }
 
+        approval_free (approval);
         return;
     }
     else if (approval != NULL && approval->type == APPROVAL_TYPE_HANDLE_WITH)
@@ -836,6 +861,7 @@ _mcd_dispatch_operation_finish (McdDispatchOperation *operation,
                          * it's OK to not distinguish */
                         tp_svc_channel_dispatch_operation_return_from_handle_with (
                             approval->context);
+                        approval->context = NULL;
                     }
                     else
                     {
@@ -844,6 +870,7 @@ _mcd_dispatch_operation_finish (McdDispatchOperation *operation,
                                successful_handler);
                         dbus_g_method_return_error (approval->context,
                                                     priv->result);
+                        approval->context = NULL;
                     }
                 }
                 else
@@ -855,13 +882,18 @@ _mcd_dispatch_operation_finish (McdDispatchOperation *operation,
                            g_quark_to_string (priv->result->domain),
                            priv->result->code, priv->result->message);
                     dbus_g_method_return_error (approval->context, priv->result);
+                    approval->context = NULL;
                 }
 
                 break;
 
             default:
-                {} /* do nothing */
+                {   /* there shouldn't be a dbus context for these: */
+                    g_assert (approval->context == NULL);
+                }
         }
+
+        approval_free (approval);
     }
 
     if (mcd_dispatch_operation_may_signal_finished (operation))
@@ -917,27 +949,97 @@ dispatch_operation_handle_with (TpSvcChannelDispatchOperation *cdo,
                                          context);
 }
 
+typedef struct {
+    McdDispatchOperation *self;
+    DBusGMethodInvocation *context;
+    gsize handler_suitable_pending;
+} ClaimAttempt;
+
+static void
+claim_attempt_resolve (ClaimAttempt *claim_attempt)
+{
+    if (claim_attempt->context != NULL)
+    {
+        g_queue_push_tail (claim_attempt->self->priv->approvals,
+                           approval_new_claim (claim_attempt->context));
+        _mcd_dispatch_operation_check_client_locks (claim_attempt->self);
+    }
+
+    g_object_unref (claim_attempt->self);
+    g_slice_free (ClaimAttempt, claim_attempt);
+}
+
+static void
+claim_attempt_suitability_cb (GObject *source,
+                              GAsyncResult *res,
+                              gpointer user_data)
+{
+    ClaimAttempt *claim_attempt = user_data;
+    GError *error = NULL;
+
+    if (!mcp_dispatch_operation_policy_handler_is_suitable_finish (
+            MCP_DISPATCH_OPERATION_POLICY (source), res, &error))
+    {
+        if (claim_attempt->context != NULL)
+            dbus_g_method_return_error (claim_attempt->context, error);
+
+        claim_attempt->context = NULL;
+        g_error_free (error);
+    }
+
+    if (--claim_attempt->handler_suitable_pending == 0)
+    {
+        DEBUG ("all plugins have finished, resolving claim attempt");
+        claim_attempt_resolve (claim_attempt);
+    }
+}
+
 static void
 dispatch_operation_claim (TpSvcChannelDispatchOperation *cdo,
                           DBusGMethodInvocation *context)
 {
     McdDispatchOperation *self = MCD_DISPATCH_OPERATION (cdo);
-    McdDispatchOperationPrivate *priv = self->priv;
+    ClaimAttempt *claim_attempt;
+    gchar *sender = dbus_g_method_get_sender (context);
+    McpDispatchOperation *plugin_api = MCP_DISPATCH_OPERATION (
+        self->priv->plugin_api);
+    const GList *p;
 
     if (self->priv->result != NULL)
     {
-        gchar *sender = dbus_g_method_get_sender (context);
 
-        DEBUG ("Giving error to %s: %s", sender, priv->result->message);
-        dbus_g_method_return_error (context, priv->result);
-
-        g_free (sender);
-
-        return;
+        DEBUG ("Giving error to %s: %s", sender, self->priv->result->message);
+        dbus_g_method_return_error (context, self->priv->result);
+        goto finally;
     }
 
-    g_queue_push_tail (priv->approvals, approval_new_claim (context));
-    _mcd_dispatch_operation_check_client_locks (self);
+    claim_attempt = g_slice_new0 (ClaimAttempt);
+    claim_attempt->self = g_object_ref (self);
+    claim_attempt->context = context;
+    claim_attempt->handler_suitable_pending = 0;
+
+    for (p = mcp_list_objects (); p != NULL; p = g_list_next (p))
+    {
+        if (MCP_IS_DISPATCH_OPERATION_POLICY (p->data))
+        {
+            McpDispatchOperationPolicy *plugin = p->data;
+
+            DEBUG ("%s: checking policy for %s",
+                G_OBJECT_TYPE_NAME (plugin), sender);
+
+            claim_attempt->handler_suitable_pending++;
+            mcp_dispatch_operation_policy_handler_is_suitable_async (plugin,
+                    NULL, sender, plugin_api,
+                    claim_attempt_suitability_cb,
+                    claim_attempt);
+        }
+    }
+
+    if (claim_attempt->handler_suitable_pending == 0)
+        claim_attempt_resolve (claim_attempt);
+
+finally:
+    g_free (sender);
 }
 
 static void
@@ -982,12 +1084,6 @@ mcd_dispatch_operation_constructor (GType type, guint n_params,
     if (!priv->client_registry || !priv->handler_map)
         goto error;
 
-    if (priv->possible_handlers == NULL && !priv->observe_only)
-    {
-        g_critical ("!observe_only => possible_handlers must not be NULL");
-        goto error;
-    }
-
     if (priv->needs_approval && priv->observe_only)
     {
         g_critical ("observe_only => needs_approval must not be TRUE");
@@ -1357,9 +1453,6 @@ _mcd_dispatch_operation_new (McdClientRegistry *client_registry,
 {
     gpointer *obj;
 
-    /* possible-handlers is only allowed to be NULL if we're only observing */
-    g_return_val_if_fail (possible_handlers != NULL || observe_only, NULL);
-
     /* If we're only observing, then the channels were requested "behind MC's
      * back", so they can't need approval (i.e. observe_only implies
      * !needs_approval) */
@@ -1731,8 +1824,13 @@ _mcd_dispatch_operation_handlers_can_bypass_approval (
     if (_mcd_dispatch_operation_is_internal (self))
         return TRUE;
 
+    /* special case: we don't have any handlers at all, so we don't want
+     * approval - we're just going to fail */
+    if (self->priv->possible_handlers == NULL)
+        return TRUE;
+
     for (iter = self->priv->possible_handlers;
-         iter != NULL && *iter != NULL;
+         *iter != NULL;
          iter++)
     {
         McdClientProxy *handler = _mcd_client_registry_lookup (
@@ -1881,7 +1979,7 @@ _mcd_dispatch_operation_handle_channels_cb (TpClient *client,
                                         tp_proxy_get_bus_name (client));
     }
 
-    self->priv->calling_handle_channels = FALSE;
+    tp_clear_object (&self->priv->trying_handler);
     _mcd_dispatch_operation_check_client_locks (self);
 }
 
@@ -1932,8 +2030,10 @@ collect_satisfied_requests (const GList *channels,
     }
 
     satisfied_requests = g_ptr_array_sized_new (g_hash_table_size (set));
+    g_ptr_array_set_free_func (satisfied_requests, g_free);
+
     request_properties = g_hash_table_new_full (g_str_hash, g_str_equal,
-        NULL, (GDestroyNotify) g_hash_table_unref);
+        g_free, (GDestroyNotify) g_hash_table_unref);
 
     g_hash_table_iter_init (&it, set);
 
@@ -1941,12 +2041,14 @@ collect_satisfied_requests (const GList *channels,
     {
         GHashTable *props;
 
-        g_ptr_array_add (satisfied_requests, path);
+        g_ptr_array_add (satisfied_requests, g_strdup (path));
         props = _mcd_request_dup_immutable_properties (value);
         g_assert (props != NULL);
-        g_hash_table_insert (request_properties, path, props);
+        g_hash_table_insert (request_properties, g_strdup (path), props);
     }
 
+    g_hash_table_unref (set);
+
     if (paths_out != NULL)
         *paths_out = satisfied_requests;
     else
@@ -2033,8 +2135,6 @@ _mcd_dispatch_operation_run_observers (McdDispatchOperation *self)
             observe_channels_cb,
             g_object_ref (self), g_object_unref, NULL);
 
-        /* don't free the individual object paths, which are borrowed from the
-         * McdChannel objects */
         g_ptr_array_unref (satisfied_requests);
 
         _mcd_tp_channel_details_free (channels_array);
@@ -2215,14 +2315,29 @@ _mcd_dispatch_operation_run_clients (McdDispatchOperation *self)
  * Invoke the handler for the given channels.
  */
 static void
-mcd_dispatch_operation_handle_channels (McdDispatchOperation *self,
-                                        McdClientProxy *handler)
+mcd_dispatch_operation_handle_channels (McdDispatchOperation *self)
 {
     GHashTable *handler_info;
     GHashTable *request_properties;
 
-    g_assert (!self->priv->calling_handle_channels);
-    self->priv->calling_handle_channels = TRUE;
+    g_assert (self->priv->trying_handler != NULL);
+
+    if (self->priv->handler_unsuitable != NULL)
+    {
+        GError *tmp = self->priv->handler_unsuitable;
+
+        /* move the error out of the way first, in case the callback
+         * tries a different handler which will also want to check
+         * handler_unsuitable */
+        self->priv->handler_unsuitable = NULL;
+
+        _mcd_dispatch_operation_handle_channels_cb (
+            (TpClient *) self->priv->trying_handler,
+            tmp, self, NULL);
+        g_error_free (tmp);
+
+        return;
+    }
 
     handler_info = tp_asv_new (NULL, NULL);
     collect_satisfied_requests (self->priv->channels, NULL,
@@ -2231,7 +2346,7 @@ mcd_dispatch_operation_handle_channels (McdDispatchOperation *self,
         MC_HASH_TYPE_OBJECT_IMMUTABLE_PROPERTIES_MAP, request_properties);
     request_properties = NULL;
 
-    _mcd_client_proxy_handle_channels (handler,
+    _mcd_client_proxy_handle_channels (self->priv->trying_handler,
         -1, self->priv->channels, self->priv->handle_with_time,
         handler_info, _mcd_dispatch_operation_handle_channels_cb,
         g_object_ref (self), g_object_unref, NULL);
@@ -2239,6 +2354,76 @@ mcd_dispatch_operation_handle_channels (McdDispatchOperation *self,
     g_hash_table_unref (handler_info);
 }
 
+static void
+mcd_dispatch_operation_handler_decision_cb (GObject *source,
+                                            GAsyncResult *res,
+                                            gpointer user_data)
+{
+    McdDispatchOperation *self = user_data;
+    GError *error = NULL;
+
+    if (!mcp_dispatch_operation_policy_handler_is_suitable_finish (
+            MCP_DISPATCH_OPERATION_POLICY (source), res, &error))
+    {
+        /* ignore any errors after the first */
+        if (self->priv->handler_unsuitable == NULL)
+            g_propagate_error (&self->priv->handler_unsuitable, error);
+        else
+            g_error_free (error);
+    }
+
+    if (--self->priv->handler_suitable_pending == 0)
+    {
+        mcd_dispatch_operation_handle_channels (self);
+    }
+
+    g_object_unref (self);
+}
+
+static void
+mcd_dispatch_operation_try_handler (McdDispatchOperation *self,
+                                    McdClientProxy *handler)
+{
+    TpClient *handler_client = (TpClient *) handler;
+    const GList *p;
+    McpDispatchOperation *plugin_api = MCP_DISPATCH_OPERATION (
+        self->priv->plugin_api);
+
+    g_assert (self->priv->trying_handler == NULL);
+    self->priv->trying_handler = g_object_ref (handler);
+
+    self->priv->handler_suitable_pending = 0;
+
+    DEBUG ("%s: channel ACL verification [%u channels]",
+        self->priv->unique_name,
+        g_list_length (self->priv->channels));
+
+    for (p = mcp_list_objects (); p != NULL; p = g_list_next (p))
+    {
+        if (MCP_IS_DISPATCH_OPERATION_POLICY (p->data))
+        {
+            McpDispatchOperationPolicy *plugin = p->data;
+
+            DEBUG ("%s: checking policy for %s",
+                G_OBJECT_TYPE_NAME (plugin),
+                tp_proxy_get_object_path (handler));
+
+            self->priv->handler_suitable_pending++;
+            mcp_dispatch_operation_policy_handler_is_suitable_async (plugin,
+                    handler_client,
+                    _mcd_client_proxy_get_unique_name (handler),
+                    plugin_api,
+                    mcd_dispatch_operation_handler_decision_cb,
+                    g_object_ref (self));
+        }
+    }
+
+    if (self->priv->handler_suitable_pending == 0)
+    {
+        mcd_dispatch_operation_handle_channels (self);
+    }
+}
+
 static gboolean
 _mcd_dispatch_operation_try_next_handler (McdDispatchOperation *self)
 {
@@ -2269,7 +2454,7 @@ _mcd_dispatch_operation_try_next_handler (McdDispatchOperation *self)
         if (handler != NULL &&
             (approval->type == APPROVAL_TYPE_HANDLE_WITH || !failed))
         {
-            mcd_dispatch_operation_handle_channels (self, handler);
+            mcd_dispatch_operation_try_handler (self, handler);
             return TRUE;
         }
 
@@ -2278,11 +2463,14 @@ _mcd_dispatch_operation_try_next_handler (McdDispatchOperation *self)
          * can legitimately try more handlers. */
         if (approval->type == APPROVAL_TYPE_HANDLE_WITH)
         {
-            GError gone = { TP_ERRORS, TP_ERROR_NOT_IMPLEMENTED,
+            GError gone = { TP_ERRORS,
+                TP_ERROR_NOT_IMPLEMENTED,
                 "The requested Handler does not exist" };
 
             g_queue_pop_head (self->priv->approvals);
+
             dbus_g_method_return_error (approval->context, &gone);
+
             approval->context = NULL;
             approval_free (approval);
             return TRUE;
@@ -2304,7 +2492,7 @@ _mcd_dispatch_operation_try_next_handler (McdDispatchOperation *self)
         if (handler != NULL && !failed &&
             (is_approved || _mcd_client_proxy_get_bypass_approval (handler)))
         {
-            mcd_dispatch_operation_handle_channels (self, handler);
+            mcd_dispatch_operation_try_handler (self, handler);
             return TRUE;
         }
     }
@@ -2439,3 +2627,11 @@ _mcd_dispatch_operation_destroy_channels (McdDispatchOperation *self)
 
     _mcd_dispatch_operation_forget_channels (self);
 }
+
+/* This should really be called ..._has_invoked_observers_if_needed,
+ * but that name would be ridiculous. */
+gboolean
+_mcd_dispatch_operation_has_invoked_observers (McdDispatchOperation *self)
+{
+    return self->priv->invoked_observers_if_needed;
+}
