@@ -415,7 +415,6 @@ _mcd_dispatcher_enter_state_machine (McdDispatcher *dispatcher,
     g_return_if_fail (channels != NULL);
     g_return_if_fail (MCD_IS_CHANNEL (channels->data));
     g_return_if_fail (requested || !only_observe);
-    g_return_if_fail (possible_handlers != NULL || only_observe);
 
     account = mcd_channel_get_account (channels->data);
     if (G_UNLIKELY (!account))
@@ -664,10 +663,10 @@ static void
 mcd_dispatcher_client_needs_recovery_cb (McdClientProxy *client,
                                          McdDispatcher *self)
 {
-    GList *channels =
+    const GList *channels =
         _mcd_handler_map_get_handled_channels (self->priv->handler_map);
     const GList *observer_filters;
-    GList *list;
+    const GList *list;
 
     DEBUG ("called");
 
@@ -706,6 +705,33 @@ mcd_dispatcher_client_needs_recovery_cb (McdClientProxy *client,
         }
 
     }
+
+    /* we also need to think about channels that are still being dispatched,
+     * but have got far enough that this client wouldn't otherwise see them */
+    for (list = self->priv->operations; list != NULL; list = list->next)
+    {
+        McdDispatchOperation *op = list->data;
+
+        if (_mcd_dispatch_operation_has_invoked_observers (op))
+        {
+            for (channels = _mcd_dispatch_operation_peek_channels (op);
+                 channels != NULL;
+                 channels = channels->next)
+            {
+                McdChannel *mcd_channel = channels->data;
+                GHashTable *properties =
+                    _mcd_channel_get_immutable_properties (mcd_channel);
+
+                if (_mcd_client_match_filters (properties, observer_filters,
+                        FALSE))
+                {
+                    _mcd_client_recover_observer (client,
+                        mcd_channel_get_tp_channel (mcd_channel),
+                        _mcd_dispatch_operation_get_account_path (op));
+                }
+            }
+        }
+    }
 }
 
 static void
@@ -1408,9 +1434,8 @@ _mcd_dispatcher_take_channels (McdDispatcher *dispatcher, GList *channels,
     {
         if (channels->next == NULL)
         {
-            DEBUG ("One channel, which cannot be handled");
-            _mcd_channel_undispatchable (channels->data);
-            g_list_free (channels);
+            DEBUG ("One channel, which cannot be handled - making a CDO "
+                   "anyway, to get Observers run");
         }
         else
         {
@@ -1424,6 +1449,8 @@ _mcd_dispatcher_take_channels (McdDispatcher *dispatcher, GList *channels,
                 _mcd_dispatcher_take_channels (dispatcher, list, requested,
                                                FALSE);
             }
+
+            return;
         }
     }
     else
@@ -1431,16 +1458,16 @@ _mcd_dispatcher_take_channels (McdDispatcher *dispatcher, GList *channels,
         DEBUG ("%s handler(s) found, dispatching %u channels",
                internal_request ? "internal" : "possible",
                g_list_length (channels));
-
-        for (list = channels; list != NULL; list = list->next)
-            _mcd_channel_set_status (MCD_CHANNEL (list->data),
-                                     MCD_CHANNEL_STATUS_DISPATCHING);
-
-        _mcd_dispatcher_enter_state_machine (dispatcher, channels,
-            (const gchar * const *) possible_handlers, requested, FALSE);
-        g_list_free (channels);
     }
 
+    for (list = channels; list != NULL; list = list->next)
+        _mcd_channel_set_status (MCD_CHANNEL (list->data),
+                                 MCD_CHANNEL_STATUS_DISPATCHING);
+
+    _mcd_dispatcher_enter_state_machine (dispatcher, channels,
+        (const gchar * const *) possible_handlers, requested, FALSE);
+    g_list_free (channels);
+
     g_strfreev (possible_handlers);
 }
 
@@ -1651,6 +1678,8 @@ _mcd_dispatcher_add_channel_request (McdDispatcher *dispatcher,
                                      McdChannel *channel, McdChannel *request)
 {
     McdChannelStatus status;
+    McdRequest *origin = _mcd_channel_get_request (request);
+    gboolean internal = _mcd_request_is_internal (origin);
 
     status = mcd_channel_get_status (channel);
 
@@ -1658,29 +1687,32 @@ _mcd_dispatcher_add_channel_request (McdDispatcher *dispatcher,
      * is not, @request must mirror the status of @channel */
     if (status == MCD_CHANNEL_STATUS_DISPATCHED)
     {
-        McdRequest *origin = _mcd_channel_get_request (request);
-        DEBUG ("reinvoking handler on channel %p", channel);
+        DEBUG ("reinvoking handler on channel %p", request);
 
         /* copy the object path and the immutable properties from the
          * existing channel */
         _mcd_channel_copy_details (request, channel);
 
-        DEBUG ("checking if ensured channel should be handled internaly");
-        if (_mcd_request_is_internal (origin))
+        if (internal)
           _mcd_request_handle_internally (origin, request, FALSE);
         else
           _mcd_dispatcher_reinvoke_handler (dispatcher, request);
     }
     else
     {
-        const gchar *preferred_handler =
-            _mcd_channel_get_request_preferred_handler (request);
-
+        DEBUG ("non-reinvoked handling of channel %p", request);
         _mcd_channel_set_request_proxy (request, channel);
-        if (status == MCD_CHANNEL_STATUS_DISPATCHING)
+
+        if (internal)
+        {
+            _mcd_request_handle_internally (origin, request, FALSE);
+        }
+        else if (status == MCD_CHANNEL_STATUS_DISPATCHING)
         {
             McdDispatchOperation *op = find_operation_from_channel (dispatcher,
                                                                     channel);
+            const gchar *preferred_handler =
+              _mcd_channel_get_request_preferred_handler (request);
 
             g_return_if_fail (op != NULL);
 
@@ -1933,6 +1965,7 @@ dispatcher_channel_request_acl_start (McdDispatcher *dispatcher,
                                    crd,
                                    dispatcher_channel_request_acl_cleanup);
 
+    g_hash_table_unref (params);
 }
 
 static void
@@ -2263,7 +2296,7 @@ send_message_got_channel (McdRequest *request,
 
             _mcd_request_unblock_account (message->account_path);
             message_context_return_error (message, error);
-            message_context_free (message);
+            _mcd_request_clear_internal_handler (request);
             g_error_free (error);
         }
     }
