@@ -38,8 +38,6 @@
  * dbus interface.
  */
 
-#include "config.h"
-
 #include <dbus/dbus.h>
 #include <string.h>
 #include <dlfcn.h>
@@ -52,10 +50,16 @@
 #include <dbus/dbus.h>
 #include <telepathy-glib/telepathy-glib.h>
 
+#include "mcd-signals-marshal.h"
+#include "mcd-dispatcher.h"
+#include "mcd-dispatcher-context.h"
+#include "mcd-account-compat.h"
 #include "mcd-connection.h"
 #include "mcd-misc.h"
 #include "mcd-service.h"
 
+#include <libmcclient/mc-errors.h>
+
 /* DBus service specifics */
 #define MISSION_CONTROL_DBUS_SERVICE "org.freedesktop.Telepathy.MissionControl5"
 
@@ -97,7 +101,7 @@ static void
 mcd_service_disconnect (McdMission *mission)
 {
     MCD_MISSION_CLASS (mcd_service_parent_class)->disconnect (mission);
-    mcd_master_shutdown (MCD_MASTER (mission), "Disconnected");
+    mcd_controller_shutdown (MCD_CONTROLLER (mission), "Disconnected");
 }
 
 static void
@@ -166,17 +170,19 @@ McdService *
 mcd_service_new (void)
 {
     McdService *obj;
+    DBusGConnection *dbus_connection;
     TpDBusDaemon *dbus_daemon;
     GError *error = NULL;
 
     /* Initialize DBus connection */
-    dbus_daemon = tp_dbus_daemon_dup (&error);
-    if (dbus_daemon == NULL)
+    dbus_connection = dbus_g_bus_get (DBUS_BUS_STARTER, &error);
+    if (dbus_connection == NULL)
     {
 	g_printerr ("Failed to open connection to bus: %s", error->message);
 	g_error_free (error);
 	return NULL;
     }
+    dbus_daemon = tp_dbus_daemon_new (dbus_connection);
     obj = g_object_new (MCD_TYPE_SERVICE,
 			"dbus-daemon", dbus_daemon,
 			NULL);
@@ -190,9 +196,3 @@ mcd_service_run (McdService * self)
     g_main_loop_run (self->main_loop);
 }
 
-void
-mcd_service_stop (McdService * self)
-{
-    if (self->main_loop != NULL)
-        g_main_loop_quit (self->main_loop);
-}
